#include <algorithm>
#include <array>
#include <cerrno>
#include <chrono>
#include <cstdio>
#include <cstring>
#include <deque>
#include <fcntl.h>
#include <fstream>
#include <functional>
#include <iomanip>
#include <iostream>
#include <limits>
#include <memory>
#include <optional>
#include <random>
#include <sstream>
#include <string>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <vector>

// const std::string text{
// "Lorem ipsum dolor sit amet, te liber verear epicurei his, vivendum pertinax
// hendrerit an pro." "Meliore graecis philosophia eu pri, intellegam cotidieque
// vis no." "Doming graecis minimum ne nec. At eam errem invidunt verterem, sit
// in noster intellegat. Te admodum similique sed."
// };

// void cout_test()
// {
// std::cout << text;
// }

// void fputs_test()
// {
// std::fputs(text.c_str(), stdout);
// }

namespace fakestd = std;

namespace detail {

template <typename C, typename = decltype(std::declval<C&>().resize(0))>
constexpr bool calc_resizable() noexcept {
	return true;
}

template <typename C, int = std::tuple_size<C>::value>
constexpr bool calc_resizable() noexcept {
	return false;
}

constexpr bool calc_resizable(...) noexcept { return false; }

// Note that when a type that is not resizable, but also doesn't have a
// constexpr size, is passed, there is a hard error.
template <typename C>
struct is_resizable {
	constexpr static bool value = calc_resizable<C>();
};
/**
 * True if and only if C is a resizable container.
 */
template <typename C>
constexpr bool is_resizable_v = is_resizable<C>::value;

template <typename C, typename = fakestd::void_t<>>
struct has_reserve {
	constexpr static bool value = false;
};

template <typename C>
struct has_reserve<C,
                   fakestd::void_t<decltype(std::declval<C&>().reserve(0))>> {
	constexpr static bool value = true;
};
/**
 * @brief True if and only if C contains an accessible reserve() member.
 */
template <typename C>
constexpr bool has_reserve_v = has_reserve<C>::value;

} // namespace detail

/**
 * @brief Attempt to reserve capacity in a container. No-op if unsupported.
 *
 * @param c The container to modify.
 * @param s The requested capacity.
 */
template <typename C,
          typename std::enable_if<detail::has_reserve_v<C>, int>::type = 0>
void try_reserve(C& c, std::size_t s) noexcept(noexcept(c.reserve(s))) {
	c.reserve(s);
	return;
}

/**
 * @brief Attempt to reserve capacity in a container. No-op if unsupported.
 *
 * @param c The container to modify.
 * @param s The requested capacity.
 */
template <typename C,
          typename std::enable_if<! detail::has_reserve_v<C>, int>::type = 0>
void try_reserve(C&, std::size_t) noexcept {
	return;
}

template <typename C, typename = void>
struct is_contiguous : std::false_type {};

template <typename C>
struct is_contiguous<C, std::void_t<decltype(std::declval<C&>().data())>>
    : std::true_type {};

template <typename C>
constexpr bool is_contiguous_v = is_contiguous<C>::value;

template <typename D = std::string, typename string,
          typename std::enable_if_t<is_contiguous_v<D>, int> = 0>
D get_file_contents_kblib(const string& filename) {
	static_assert(std::is_trivially_copyable<typename D::value_type>::value,
	              "D must be a sequence of trivial types");
	static_assert(sizeof(typename D::value_type) == 1,
	              "D must be a sequence of char-sized objects.");
	std::ifstream in(filename, std::ios::in | std::ios::binary);
	if (in) {
		D contents;
		in.seekg(0, std::ios::end);
		auto size = in.tellg();
		contents.resize(size);
		in.seekg(0, std::ios::beg);
		in.read(reinterpret_cast<char*>(&contents[0]), size);
		in.close();
		return contents;
	} else {
		throw 1;
	}
}

template <typename D = std::string, typename string,
          typename std::enable_if_t<! is_contiguous_v<D>, int> = 0>
D get_file_contents_kblib(const string& filename) {
	static_assert(std::is_trivially_copyable<typename D::value_type>::value,
	              "D must be a sequence of trivial types");
	static_assert(sizeof(typename D::value_type) == 1,
	              "D must be a sequence of char-sized objects.");
	std::ifstream in(filename, std::ios::in | std::ios::binary);
	if (in) {
		D contents;
		in.seekg(0, std::ios::end);
		try_reserve(contents, in.tellg());
		in.seekg(0, std::ios::beg);
		std::copy((std::istreambuf_iterator<char>(in)),
		          std::istreambuf_iterator<char>(), std::back_inserter(contents));
		in.close();
		return contents;
	} else {
		throw 1;
	}
}

template <typename D, typename string>
D get_file_contents_c(const string& filename) {
	std::FILE* fp = std::fopen(filename, "rb");
	if (fp) {
		D contents;
		std::fseek(fp, 0, SEEK_END);
		contents.resize(std::ftell(fp));
		std::rewind(fp);
		std::fread(&contents[0], 1, contents.size(), fp);
		std::fclose(fp);
		return contents;
	}
	throw errno;
}

template <typename D, typename string>
D get_file_contents_cpp(const string& filename) {
	std::ifstream in(filename, std::ios::in | std::ios::binary);
	if (in) {
		D contents;
		in.seekg(0, std::ios::end);
		contents.resize(in.tellg());
		in.seekg(0, std::ios::beg);
		in.read(&contents[0], contents.size());
		in.close();
		return contents;
	}
	throw errno;
}

template <typename D, typename string>
D get_file_contents_it(const string& filename) {
	std::ifstream in(filename, std::ios::in | std::ios::binary);
	if (in) {
		return {(std::istreambuf_iterator<char>(in)),
		        std::istreambuf_iterator<char>()};
	}
	throw errno;
}

template <typename D, typename string>
D get_file_contents_assign(const string& filename) {
	std::ifstream in(filename, std::ios::in | std::ios::binary);
	if (in) {
		D contents;
		in.seekg(0, std::ios::end);
		contents.reserve(in.tellg());
		in.seekg(0, std::ios::beg);
		contents.assign((std::istreambuf_iterator<char>(in)),
		                std::istreambuf_iterator<char>());
		in.close();
		return contents;
	}
	throw errno;
}

template <typename D, typename string>
D get_file_contents_copy(const string& filename) {
	std::ifstream in(filename, std::ios::in | std::ios::binary);
	if (in) {
		D contents;
		in.seekg(0, std::ios::end);
		contents.reserve(in.tellg());
		in.seekg(0, std::ios::beg);
		std::copy((std::istreambuf_iterator<char>(in)),
		          std::istreambuf_iterator<char>(), std::back_inserter(contents));
		in.close();
		return contents;
	}
	throw errno;
}

template <typename D, typename string, int block = 1000>
D get_file_contents_me(const string& filename) {
	std::ifstream in{};
	in.exceptions(std::ios::badbit);
	in.open(filename, std::ios::in | std::ios::binary);
	D contents;
	alignas(16) std::array<char, block> buf;
	while (in) {
		in.read(buf.begin(), buf.size());
		contents.insert(contents.end(), buf.begin(), buf.begin() + in.gcount());
	}
	return contents;
}

template <typename string>
std::string get_file_contents_rdbuf(const string& filename) {
	std::ifstream in(filename, std::ios::in | std::ios::binary);
	if (in) {
		std::ostringstream contents;
		contents << in.rdbuf();
		in.close();
		return contents.str();
	}
	throw errno;
}

template <typename string, int block = 1000>
inline std::string get_file_contents_append(const string& filename) {
	std::ifstream in{};
	in.exceptions(std::ios::badbit);
	in.open(filename, std::ios::in | std::ios::binary);
	std::string contents;
	std::array<char, block> buf;
	while (in) {
		in.read(buf.begin(), buf.size());
		contents.append(buf.begin(), in.gcount());
	}
	return contents;
}

/*
template <typename D, typename string>
std::optional<D> try_get_file_contents_mmap(const string& filename) {
   auto fd = open(std::data(filename), O_RDONLY);
   if (fd == -1) {
      perror("open");
      return std::nullopt;
   }
   struct stat sb;
   if (fstat(fd, &sb) == -1) {
      perror("fstat");
      return std::nullopt;
   }
   auto p = mmap(nullptr, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
   if (p == MAP_FAILED) {
      perror("mmap");
      return std::nullopt;
   }
   std::optional<D> r{std::in_place, p, p + sb.st_size};
   munmap(p,sb.st_size);
   return r;
}
//*/

auto file_size(const char* filename) -> std::size_t {
	struct stat sb;
	if (stat(filename, &sb)) {
		perror("stat");
		return 0;
	}
	return sb.st_size;
}

template <typename D>
D get_file_contents_mmap_orig(const char* filename) {
	static_assert(std::is_trivially_copyable<typename D::value_type>::value,
	              "D must be a sequence of trivial types");
	static_assert(sizeof(typename D::value_type) == 1,
	              "D must be a sequence of char-sized objects.");
	auto fd = open(filename, O_RDONLY);
	if (fd == -1) {
		perror("open");
		throw 0;
	}
	struct stat sb;
	if (fstat(fd, &sb) == -1) {
		perror("fstat");
		throw 0;
	}
	auto unmapper = [length = sb.st_size](void* p) { munmap(p, length); };
	auto p = std::unique_ptr<void, decltype(unmapper)>(
	    mmap(nullptr, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0), unmapper);
	close(fd);
	if (p.get() == MAP_FAILED) {
		perror("mmap");
		throw 0;
	}
	D r;
	r.resize(sb.st_size);
	std::memcpy(std::data(r), p.get(), sb.st_size);
	return r;
}
template <typename D>
D get_file_contents_mmap(const char* filename) {
	static_assert(std::is_trivially_copyable<typename D::value_type>::value,
	              "D must be a sequence of trivial types");
	static_assert(sizeof(typename D::value_type) == 1,
	              "D must be a sequence of char-sized objects.");
	auto fd = open(filename, O_RDONLY);
	if (fd == -1) {
		perror("open");
		throw 0;
	}
	struct stat sb;
	if (fstat(fd, &sb) == -1) {
		perror("fstat");
		throw 0;
	}
	auto unmapper = [length = sb.st_size](void* p) { munmap(p, length); };
	auto p = std::unique_ptr<void, decltype(unmapper)>(
	    mmap(nullptr, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0), unmapper);
	close(fd);
	if (p.get() == MAP_FAILED) {
		perror("mmap");
		throw 0;
	}
	auto q = static_cast<typename D::pointer>(p.get());
	return D{q, q + sb.st_size};
}

template <typename T>
auto do_time(T func, int runs) {
	using namespace std::chrono;

	auto start{high_resolution_clock::now()};
	for (auto i{0}; i < runs; ++i) {
		func();
	}
	auto end{high_resolution_clock::now()};

	return duration_cast<milliseconds>(end - start).count();
}

using time_c = decltype(std::chrono::milliseconds{}.count());
using time_f = std::function<time_c()>;

template <typename T>
auto time(T func, int runs) -> time_f {
	return [=] { return do_time(func, runs); };
}

volatile long long bytes_read{0};

template <typename F, typename I>
auto test_method(F f, I i) {
	auto s = f(i);
	bytes_read += s.size();
	return s.size();
}

int io_main(int argc, char** argv) {
	using namespace std::chrono;
	using namespace std::literals;

	auto use_stderr = false;

	// convenient noop for decltype'ing
	void _();
	using vec = std::vector<char>;
	using bvec = std::vector<std::byte>;
	using uvec = std::vector<unsigned char>;
	using str = std::string;

	using label_t = const char*;

	auto setup_test = [](auto runs, auto filename, auto& times) {
		// get the file in cache
		do_time(
		    [=] {
			    test_method(get_file_contents_cpp<str, decltype(filename)>,
			                filename);
		    },
		    2);

#define TEST_METHOD(...) time([=] { test_method(__VA_ARGS__, filename); }, runs)

		if (true) {
			times.emplace_back(
			    "C++\tstr",
			    TEST_METHOD(get_file_contents_cpp<str, decltype(filename)>));
			times.emplace_back(
			    "C++\tcvec",
			    TEST_METHOD(get_file_contents_cpp<vec, decltype(filename)>));
			// times.emplace_back("C++\tbvec",
			// TEST_METHOD(get_file_contents_cpp<bvec,
			// decltype(filename)>));
			// times.emplace_back("C++\tuvec",
			// TEST_METHOD(get_file_contents_cpp<uvec,
			// decltype(filename)>));
			times.emplace_back(
			    "C\tstr",
			    TEST_METHOD(get_file_contents_c<str, decltype(filename)>));
			times.emplace_back(
			    "C\tcvec",
			    TEST_METHOD(get_file_contents_c<vec, decltype(filename)>));
			times.emplace_back(
			    "C\tbvec",
			    TEST_METHOD(get_file_contents_c<bvec, decltype(filename)>));
			times.emplace_back(
			    "C\tuvec",
			    TEST_METHOD(get_file_contents_c<uvec, decltype(filename)>));
			// works but too slow (1.8 s / GiB)
			// times.emplace_back("buf\tstr",
			// TEST_METHOD(get_file_contents_rdbuf<decltype(filename)>));
		}

		if (true) {
			times.emplace_back(
			    "kblib\tstr",
			    TEST_METHOD(get_file_contents_kblib<str, decltype(filename)>));
			times.emplace_back(
			    "kblib\tcvec",
			    TEST_METHOD(get_file_contents_kblib<vec, decltype(filename)>));
			times.emplace_back(
			    "kblib\tbvec",
			    TEST_METHOD(get_file_contents_kblib<bvec, decltype(filename)>));
			times.emplace_back(
			    "kblib\tuvec",
			    TEST_METHOD(get_file_contents_kblib<uvec, decltype(filename)>));
			// works but too slow for testing (5.6 s / GiB)
			// times.emplace_back("kblib\tdeque",
			// TEST_METHOD(get_file_contents_kblib<std::deque<char>,
			// decltype(filename)>));
		}
		if (true) {
			times.emplace_back("mmap\tstr",
			                   TEST_METHOD(get_file_contents_mmap<str>));
			times.emplace_back("mmap\tcvec",
			                   TEST_METHOD(get_file_contents_mmap<vec>));
			times.emplace_back("mmap\tbvec",
			                   TEST_METHOD(get_file_contents_mmap<bvec>));
			times.emplace_back("mmap\tuvec",
			                   TEST_METHOD(get_file_contents_mmap<uvec>));
			times.emplace_back("mmap_copy\tstr",
			                   TEST_METHOD(get_file_contents_mmap_orig<str>));
			times.emplace_back("mmap_copy\tcvec",
			                   TEST_METHOD(get_file_contents_mmap_orig<vec>));
			times.emplace_back("mmap_copy\tbvec",
			                   TEST_METHOD(get_file_contents_mmap_orig<bvec>));
			times.emplace_back("mmap_copy\tuvec",
			                   TEST_METHOD(get_file_contents_mmap_orig<uvec>));
		}
		if (false) {
			times.emplace_back(
			    "it\tstr",
			    TEST_METHOD(get_file_contents_it<str, decltype(filename)>));
			times.emplace_back(
			    "it\tcvec",
			    TEST_METHOD(get_file_contents_it<vec, decltype(filename)>));
			// times.emplace_back("it\tbvec",
			// TEST_METHOD(get_file_contents_it<bvec, decltype(filename)>));
			times.emplace_back(
			    "it\tuvec",
			    TEST_METHOD(get_file_contents_it<uvec, decltype(filename)>));
		}
		if (false) {
			times.emplace_back(
			    "assign\tstr",
			    TEST_METHOD(get_file_contents_assign<str, decltype(filename)>));
			times.emplace_back(
			    "assign\tcvec",
			    TEST_METHOD(get_file_contents_assign<vec, decltype(filename)>));
			// times.emplace_back("assign\tbvec",
			// TEST_METHOD(get_file_contents_assign<bvec,
			// decltype(filename)>));
			times.emplace_back(
			    "assign\tuvec",
			    TEST_METHOD(get_file_contents_assign<uvec, decltype(filename)>));
			times.emplace_back(
			    "copy\tstr",
			    TEST_METHOD(get_file_contents_copy<str, decltype(filename)>));
			times.emplace_back(
			    "copy\tcvec",
			    TEST_METHOD(get_file_contents_copy<vec, decltype(filename)>));
			// times.emplace_back("copy\tbvec",
			// TEST_METHOD(get_file_contents_copy<bvec,
			// decltype(filename)>));
			times.emplace_back(
			    "copy\tuvec",
			    TEST_METHOD(get_file_contents_copy<uvec, decltype(filename)>));
		}
		if (false) {
			times.emplace_back(
			    "me 256\tstr",
			    TEST_METHOD(get_file_contents_me<str, decltype(filename), 256>));
			times.emplace_back(
			    "me 256\tcvec",
			    TEST_METHOD(get_file_contents_me<vec, decltype(filename), 256>));
			// times.emplace_back("me 256\tbvec",
			// TEST_METHOD(get_file_contents_me<bvec, decltype(filename),
			// 256>));
			times.emplace_back(
			    "me 256\tuvec",
			    TEST_METHOD(get_file_contents_me<uvec, decltype(filename), 256>));
			times.emplace_back(
			    "me 40960\tstr",
			    TEST_METHOD(get_file_contents_me<str, decltype(filename), 40960>));
			times.emplace_back(
			    "me 40960\tcvec",
			    TEST_METHOD(get_file_contents_me<vec, decltype(filename), 40960>));
			// times.emplace_back("me 40960\tbvec",
			// TEST_METHOD(get_file_contents_me<bvec, decltype(filename),
			// 40960>));
			times.emplace_back(
			    "me 40960\tuvec",
			    TEST_METHOD(
			        get_file_contents_me<uvec, decltype(filename), 40960>));
		}
		if (false) {
			times.emplace_back(
			    "me 1000\tstr",
			    TEST_METHOD(get_file_contents_me<str, decltype(filename), 1000>));
			times.emplace_back(
			    "me 1000\tcvec",
			    TEST_METHOD(get_file_contents_me<vec, decltype(filename), 1000>));
			times.emplace_back(
			    "me 1024\tstr",
			    TEST_METHOD(get_file_contents_me<str, decltype(filename), 1024>));
			times.emplace_back(
			    "me 1024\tcvec",
			    TEST_METHOD(get_file_contents_me<vec, decltype(filename), 1024>));
			times.emplace_back(
			    "me 1048576\tstr",
			    TEST_METHOD(
			        get_file_contents_me<str, decltype(filename), 1048576>));
			times.emplace_back(
			    "me 1048576\tcvec",
			    TEST_METHOD(
			        get_file_contents_me<vec, decltype(filename), 1048576>));
			times.emplace_back(
			    "me 20000\tstr",
			    TEST_METHOD(get_file_contents_me<str, decltype(filename), 20000>));
			times.emplace_back(
			    "me 20000\tcvec",
			    TEST_METHOD(get_file_contents_me<vec, decltype(filename), 20000>));
			times.emplace_back(
			    "me 10240\tstr",
			    TEST_METHOD(get_file_contents_me<str, decltype(filename), 10240>));
			times.emplace_back(
			    "me 10240\tcvec",
			    TEST_METHOD(get_file_contents_me<vec, decltype(filename), 10240>));
			times.emplace_back(
			    "me 40000\tstr",
			    TEST_METHOD(get_file_contents_me<str, decltype(filename), 40000>));
			times.emplace_back(
			    "me 40000\tcvec",
			    TEST_METHOD(get_file_contents_me<vec, decltype(filename), 40000>));

			times.emplace_back(
			    "me 32768\tstr",
			    TEST_METHOD(get_file_contents_me<str, decltype(filename), 32768>));
			times.emplace_back(
			    "me 32768\tcvec",
			    TEST_METHOD(get_file_contents_me<vec, decltype(filename), 32768>));
			times.emplace_back(
			    "me 20480\tstr",
			    TEST_METHOD(get_file_contents_me<str, decltype(filename), 20480>));
			times.emplace_back(
			    "me 20480\tcvec",
			    TEST_METHOD(get_file_contents_me<vec, decltype(filename), 20480>));
		}
	};

	struct Row {
		label_t label;
		time_c time;
		float speed;
		float percent_best;
	};

	auto prepare_results = [](const auto& times, float GiBs) {
		std::vector<Row> results;
		float best_speed = 0;
		std::transform(times.begin(), times.end(), std::back_inserter(results),
		               [=, &best_speed](auto p) {
			               const auto c = p.second();
			               const auto s = (GiBs / c) * 1000;
			               best_speed = std::max(best_speed, s);
			               return Row{p.first, c, s, 1};
		               });
#if 1
		std::stable_sort(results.begin(), results.end(),
		                 [](auto a, auto b) { return a.time < b.time; });
		// float best = (-1u) / results.front().speed;
#else
		std::stable_sort(results.begin(), results.end(), [](auto a, auto b) {
			return std::string_view{a.label} < b.label;
		});
#endif
		for (auto& row : results) {
			row.percent_best = 100 * row.speed / best_speed;
		}
		return results;
	};

	auto& print = use_stderr ? std::clog : std::cout;
	bool quiet{};

	auto show_times = [&](auto filename, std::size_t target_bytes,
	                      std::size_t filesize) {
		int runs = target_bytes / filesize;
		const auto size_kb = static_cast<int>(std::ceil(filesize / 1024.f));

		std::vector<std::pair<label_t, time_f>> times;
		setup_test(runs, filename, times);
		std::shuffle(times.begin(), times.end(), std::random_device{});

		if (not quiet) {
			print << filename << ": ";
		}
		print << size_kb << " KiB   ";

		filesize *= runs;
		const auto total_GiBs = (float)filesize / (1024 * 1024 * 1024);

		if (not quiet) {
			print << "\nRead " << std::setprecision(2) << std::fixed
			      << (total_GiBs * 1024) << " MiB total (" << runs
			      << " iterations) in:\nmethod\ttype";
		} else {
			print << "\ttype";
		}
		print << "\tms\tGiB/s\t%best" << std::endl;

		auto results = prepare_results(times, total_GiBs);

		for (auto row : results) {
			print << row.label << '\t' << row.time << '\t' << std::setprecision(2)
			      << std::fixed << row.speed << '\t' << std::setprecision(2)
			      << std::fixed << row.percent_best << "%\n";
			// print<<std::setw(16)<<std::left<<i.first<<' '<<i.second<<'\t'
			// <<std::setprecision(2)<<std::fixed
			// <<((float)filesize/file_units)/i.second<<'\n';
		}
		print << std::endl;
	};

	constexpr std::size_t target_bytes = 1024 * 1024 * 1024; // 1 GiB

	if (argc > 1 and argv[1] == "-q"sv) {
		quiet = true;
		--argc;
		++argv;
	}

	if (argc > 1) {
		if (not quiet) {
			print << "Small file:\n";
		}
		auto filesize = file_size(argv[1]);
		show_times(argv[1], target_bytes, filesize);
	}
	if (argc > 2) {
		if (not quiet) {
			print << "Medium file:\n";
		}
		auto filesize = file_size(argv[2]);
		show_times(argv[2], target_bytes, filesize);
	}
	if (argc > 3) {
		if (not quiet) {
			print << "Large file:\n";
		}
		auto filesize = file_size(argv[3]);
		show_times(argv[3], target_bytes, filesize);
	}
	if (argc > 4) {
		if (not quiet) {
			print << "Huge file:\n";
		}
		auto filesize = file_size(argv[4]);
		show_times(argv[4], target_bytes, filesize);
	}

	// auto fputs_time{ time(fputs_test, runs) };

	// auto cout_time{ time(cout_test, runs) };
	// std::ios::sync_with_stdio(false);
	// std::cout.tie(nullptr);
	// auto cout_utime{ time(cout_test, runs) };

	// std::cout << "\n";
	// print << "fputs_time " << fputs_time
	// << " cout_time " << cout_time
	// << " cout_utime " << cout_utime << std::endl;
	return 0;
}
